%%% SICStus Prolog Code for Protein Folding Prediction
%%% in CLP(FD) 
%%% Version JUNE 21th, 2004

%%% GENERAL REMARK: The output in PDB standard format
%%% is written either in the
%%% 1. SICStus console or in 
%%% 2. a file called "output-XXXX.pdb" in your
%%%    SICStus "working directory"
%%%    (choose your own working directory clicking 
%%%    on "File" options)
%%%    where XXXX is the pdb name

:-use_module(library(clpfd)).
:-use_module(library(lists)).
:-use_module(library(terms)).
:-use_module(library(timeout)).
:-use_module(library(system),[]).
:-consult('data.pl').
:-consult('table.pl').
:-consult('torsions.pl').
:-prolog_flag(unknown,_,fail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% (MAIN) PREDICATE: fcc_pf/1-2-3
%%% INPUT: ID, the code of the protein according to names in file data.pl
%%%        Time (time limit for the search in SECONDS)
%%%        A parameter that "compacts" the solution
%%% OUTPUT: the minimum Energy associated to the conformation
%%%         represented by Tertiary
%%% EXAMPLE:  fcc_pf('1ED0').
%%%         Default time limit: 24h, 
%%%         Default compact:  (4.4*N**0.38)/N
%%%           fcc_pf('1ED0',120).
%%%         user time limit: 120s, 
%%%         Default compact:  (4.4*N**0.38)/N
%%%           fcc_pf('1ED0',120,0.6).
%%%         user time limit: 120s, user compact parameter: 0.6          
%%% NOTE: "protein" is defined in data.pl
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          
fcc_pf( ID ):-
    protein(ID, Primary, _), 
    length(Primary, N),
    C is (4.4*N**0.38)/N,
    fcc_pf(ID, 36000, C).
fcc_pf( ID, Time):-
    protein(ID, Primary, _), 
    length(Primary, N),
    C is (4.4*N**0.38)/N,
    fcc_pf(ID, Time, C).
fcc_pf( ID, Time, Compact):-
    initialization,
    protein(ID, Primary, Secondary), 
    constrain(Primary, Secondary, Indexes, Tertiary, Energy, Matrix, Freq, Compact),
    writetime(video,'Constraint time: '),nl,!,
    solution_search(Time, Primary, Secondary, Indexes, 
                     Tertiary, Energy, Matrix, Freq),                 
    print_results(ID,Time,Primary, Secondary, Tertiary,Compact).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% CONSTRAINT DEFINITION PART  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain/8
%%% INPUT: The lists Primary and Secondary of the protein
%%% OUTPUT: The lists Indexes, Tertiary, 
%%%     the Energy (SERVE ???)
%%%     the potential matrix and
%%%     the Frequency list for indexes.
%%% EFFECT: add constraints to all existing variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

constrain(Primary, Secondary, Indexes, Tertiary, Energy, Matrix, Freq,Compact):-
    %%% Input test:
    checkamino(Primary), 
    %%% Basic Encoding:
    length(Primary,N),
    generate_indexes(N, Indexes),
    generate_tertiary(N, Tertiary),
    domain_bounds(N, Tertiary),
    avoid_self_loops(Tertiary,N),
    next_constraints(Tertiary),
    distance_constraints(Tertiary),
    compact_constraints(Tertiary,N,Compact),  
    %%% Other constraints:
    secondary_info(Secondary, Indexes, Tertiary),
    gettorsions(Primary, Indexes, Freq),
    avoid_symmetries(Indexes, Secondary, Tertiary),
    indexes_to_coordinates(Indexes, Tertiary),
    energy_constraints(Primary, Tertiary, Indexes, 
                       Energy, _ListaPot, Matrix).
                       
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: generate_indexes/2
%%% INPUT: N, a number greater than or equal to 0,
%%% OUTPUT: a list of N-2 variables as \theta-indexes
%%% 0-5 six torsion angles (0,60,120,180,240,300)
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_indexes(N, Indexes):-
    N > 3, !, 
    M is N-3,
    length(Indexes,M),
    domain(Indexes,0,5).
generate_indexes( _, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: generate_tertiary/2
%%% INPUT: N, the length of the primary list of aminoacids
%%% OUTPUT: Tertiary,  a list of 3*N variables to be used as coordinates
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_tertiary(N,Tertiary):-
    N3 is N*3,
    length(Tertiary,N3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: domain_bounds/2
%%% INPUT: N, a number greater than or equal to 0,
%%%        and the Tertiary list
%%% EFFECT: adding FD constraints to those variables
%%%        in order they keep value in 2*n fcc
%%% NOTE: "X + Y + Z #= Sum" would be compiled into 
%%%       internal scalar_product. sum is simpler.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

domain_bounds(N, [X,Y,Z|Rest]) :-
    CUBESIZE is N * 2,
    domain([X,Y,Z],0,CUBESIZE),
    sum([X,Y,Z], #=, Sum),
    even(Sum),
    domain_bounds(N, Rest).
domain_bounds(_, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: avoid_self_loops/2 (aux.: positions_to_integers/3)
%%% INPUT:  Tertiary, the list of variables, N size of protein
%%% EFFECT: adding FD constraints to those variables
%%%        in order to avoid equal triples
%%% NOTE1: It introduces large numbers.
%%% NOTE2: "all_different" does not propagate until labeling.
%%% status: Better abstract it by simpler constraints ???
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

avoid_self_loops(Tertiary, N):-
      BASE is 2*N+1,
      positions_to_integers(Tertiary, ListaInteri, BASE),
      all_different(ListaInteri).

positions_to_integers([X,Y,Z|R], [I|S], BASE):-
      I  #= X*BASE*BASE+Y*BASE+Z,  %%% weighted sum !!!
      positions_to_integers( R , S, BASE).
positions_to_integers([],[],_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: next_constraints/1 (aux.: next/6)
%%% INPUT: Tertiary, the list of variables
%%% EFFECT: adding FD constraints to those variables
%%%        in order to set that the triples are consecutive
%%% NOTE: It introduces NON linear constraints.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

next_constraints([X1,Y1,Z1,X2,Y2,Z2|C]) :-
    next(X1,Y1,Z1,X2,Y2,Z2),
    next_constraints([X2,Y2,Z2|C]).
next_constraints([_,_,_]).
next_constraints([]).

next(X1,Y1,Z1,X2,Y2,Z2):-
     domain([Dx,Dy,Dz],0,1),
     Dx #= abs(X1-X2),
     Dy #= abs(Y1-Y2),
     Dz #= abs(Z1-Z2),
     sum([Dx,Dy,Dz], #= ,  2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: distance_constraints/1 (aux.: distance/4, non_next/6)
%%% INPUT: Tertiary, the list of variables' triples
%%% EFFECT: adding FD constraints to those variables
%%%        in order to set that non consecutive triples
%%%        has lattice distance greater than 1.
%%% NOTE1: It introduces NON linear constraints.
%%% NOTE2: The ISO notation X ** 2 (for X^2) does
%%%        not work in clpfd
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

distance_constraints([X1,Y1,Z1,X2,Y2,Z2|R]):-
    distance(X1,Y1,Z1,R),
    distance_constraints([X2,Y2,Z2|R]).
distance_constraints([_,_,_]).
distance_constraints([]).

distance(X1,Y1,Z1,[X2,Y2,Z2|R]):-
    non_next(X1,Y1,Z1,X2,Y2,Z2),
    distance(X1,Y1,Z1,R).
distance(_,_,_,[]).

non_next(X1,Y1,Z1,X2,Y2,Z2):-  
  Dx #= (X1-X2) * (X1-X2),
  Dy #= (Y1-Y2) * (Y1-Y2),
  Dz #= (Z1-Z2) * (Z1-Z2),
  sum([Dx,Dy,Dz], #>, 2). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: compact_constraints/3 
%%% INPUT: Tertiary, the list of variables' triples and N
%%% EFFECT: adding FD constraints to those variables
%%%        in order to set that the (linear) distance between
%%%        aminoacids is small.
%%% status: ESPERIMENTO
%%% con 4*N/5 perdevo sol nella 1KVG
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

compact_constraints(Tertiary,N,Ratio):-
     Dist is integer(N*Ratio),
     dist_constraints(Tertiary,Dist).

dist_constraints([],_).
dist_constraints([_,_,_],_).
dist_constraints([X1,Y1,Z1,X2,Y2,Z2|R], Dist) :-
    dist_aux(X1,[X2,Y2,Z2|R],Dist),
    dist_aux(Y1,[Y2,Z2|R],Dist),
    dist_aux(Z1,[Z2|R],Dist),
    dist_constraints([X2,Y2,Z2|R], Dist).
dist_aux(_,[],_).
dist_aux(Z1,[Z2],Dist):-
    offset(Z1,Z2,Dist).
dist_aux(Y1,[Y2,_],Dist):- %% Y
    offset(Y1,Y2,Dist).
dist_aux(X1,[X2,_,_|R],Dist):- %%eihter X or Y or Z
    offset(X1,X2,Dist),
    dist_aux(X1,R,Dist).
offset(A,B,Dist):-  
  A - B #< Dist,
  B - A #< Dist.
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: secondary_info/3 
%%% INPUT: The list Secondary,
%%%        The list Indexes, 
%%%        The list Tertiary
%%% EFFECT: Adding constraints to the variables in Indexes and Tertiary
%%%        according to the information in Secondary
%%% NOTE: helix(1,..) or strand(1,..) are not allowed!!!
%%% NOTE: helix(_,N) or strand(_,N) are not allowed!!!
%%%       We do not add +1 because in this way we leave free
%%%       the beginning and the ending of a chain/strand
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Case 0: Empty list
secondary_info([], _ , _ ).
%%% Case 1: ssbond
secondary_info([ssbond(A1,A2)|SecRest], Indexes, Tertiary):-
    !,
    ssbond_propagate(A1,A2,Tertiary),
    secondary_info(SecRest, Indexes, Tertiary).
%%% Case 2: helix
secondary_info([helix(Start, End)|SecRest], Indexes, Tertiary):-
     !,
     S is Start - 1,
     Helix_length #= End - Start, % We do not add + 1 
     sublist(S, Helix_length, Indexes, Subindexes),
     helix(Subindexes),
     secondary_info(SecRest,Indexes,Tertiary).
%%% Case 3: strand
secondary_info([strand(Start, End)|SecRest], Indexes, Tertiary):-
     !,
     S is Start - 1,
     Helix_length #= End - Start, % We do not add + 1 
     sublist(S, Helix_length, Indexes, Subindexes),
     strand(Subindexes),
     secondary_info(SecRest,Indexes, Tertiary).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: helix/1 
%%% INPUT: A sublist of Indexes, 
%%% EFFECT: Adding constraints to the variables involved in the sublist
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

helix(Indexes):-
   domain(Indexes,5,5).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: strand/1
%%% INPUT: A sublist of Indexes, 
%%% EFFECT: Adding constraints to the variables involved in the sublist
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

strand(Indexes):-
   domain(Indexes,3,3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: ssbond_propagate/3 (aux. ssbond/2)
%%% INPUT: The number of two aminoacids (A1 and A2) and the Tertiary
%%% structure. 
%%% EFFECT: An upper bound Limit = 36 to the sum
%%%         of the square of their distances is set.
%%% NOTE: A NON linear constraint in introduced
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ssbond_propagate(A1,A2,Tertiary):-
    terelem(Tertiary,A1,P1),
    terelem(Tertiary,A2,P2),
    ssbond(P1,P2).
ssbond([X1,Y1,Z1],[X2,Y2,Z2]):-
    Dx #= (X1-X2)*(X1-X2),
    Dy #= (Y1-Y2)*(Y1-Y2),
    Dz #= (Z1-Z2)*(Z1-Z2),
%%%    sum([Dx,Dy,Dz], #=<, 36). %%%%% BOND LIMIT 
    sum([Dx,Dy,Dz], #=<, 36). 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: avoid_symmetries/3
%%% INPUT: Indexes, Secondary and Tertiary Structure
%%% EFFECT: Adds constraints to avoid symmetries.
%%%    In particular, the "best" subsequence of Indexes is 
%%%    identified (Sublist). A point "Center" in it is then selected.
%%%    Center is the starting point of the longest ground
%%%    sequence in Sublist.
%%%    The coordinates of this point and of the next,
%%%    as well as two coordinates of the sequent are set.
%%%    status: almost OK.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

avoid_symmetries(Indexes,Secondary,Tertiary):-
    length(Tertiary,N),
    N3 #= N/3,
    Odd is N3 mod 2,
    P is N3+Odd, P1 is P+1, P2 is P+2,
    %%% P and P2 are even. P1 is odd.
    choose_subindexes(Indexes, Start, Length, Secondary),
    sublist(Start, Length,Indexes,Sublist),
    getmaxrun(Sublist,Center),
    %% Center is the pointer to the first ground indexes of the maximal run (correct)
    %% S,S1,S2  are the tertiary pos assigned producing the first index, then propagation 
    %% fills the whole run
    S is Start + Center - 1,
    S1 is S+1, 
    S2 is S+2,
    terelem(Tertiary,S,[P,P,P]),
    terelem(Tertiary,S1,[P1,P1,P]),
    terelem(Tertiary,S2,[P2,P,_P3]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: gettorsions/3 (aux. get_torsions/3)
%%% INPUT: The primary list (the first aminoacid is
%%% not considered), The Indexes
%%% Output: A list Freq of pairs
%%% [ Index, [5, 3, 2, 4, 0, 1] ]
%%% where Index is a nonground variable
%%% while the numbers denote the admissible torsion angles
%%% in decreasing frequency order.
%%% state: almost OK
%%% NOTE1: La lista di frequenze e' cancellata
%%% (teniamo solo l'ordine). Cio' snellisce un po'
%%% le dim. dei dati passati. 
%%% NOTE2: ho semplificato il predicato togliendo assoc_vals!!!
%%% e le coppie in cui Index e' gia' ground.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gettorsions([_|Rest],Indexes, Freq) :-
     get_torsions(Rest,Indexes,Freq).
get_torsions([_A1,A2,A3|Primary],[I|Indexes], Freq) :-
     ground(I),!,
     get_torsions([A2,A3|Primary],Indexes, Freq).
get_torsions([A1,A2,A3|Primary],[I|Indexes],[[I,Order]|Rest]) :-
     torsion(A1,A2,Order,_Frequencies),!,
     get_torsions([A2,A3|Primary],Indexes, Rest).
get_torsions(_,_,[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: indexes_to_coordinates/2 
%%%    (aux.  indexes_to_coordinates/3, 
%%%         indexes_to_directions, directions_to_coordinates,
%%%          vector_sum)
%%% INPUT: The list of Indexes
%%%        The list of variables's triples Tertiary
%%% EFFECT: Relates the \theta-constraints on Indexes to the 
%%%        coordinates variables in Tertiary
%%% status: doubts
%%% AGO: Veniva chiamato con il terzo parametro libero.
%%% Allora l'ho sostituito con una chiamata con 2 parametri.
%%% E' strano che il primo passo sia directions_to_coordinates
%%% visto che Directions e' l'unica cosa non nota... o sbaglio.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

indexes_to_coordinates(Indexes, Tertiary) :-
       indexes_to_coordinates(Indexes, Tertiary, _Directions).
       
indexes_to_coordinates([], _, _).
indexes_to_coordinates(Indexes, Tertiary, Directions):-
    directions_to_coordinates(Directions, Tertiary),
    indexes_to_directions(Indexes, Directions).

directions_to_coordinates([D1 | Directions],[X11,X12,X13,X21,X22,X23|Coordinate]):-
    vector_sum(D1, [X11,X12,X13], [X21,X22,X23]),
    directions_to_coordinates(Directions,[X21,X22,X23|Coordinate]).
directions_to_coordinates([], [_,_,_] ).
directions_to_coordinates([], [] ).

indexes_to_directions([I1|Indexes],[D0,D1,D2|Directions]):-
    new_direction(D0,D1,D2,I1),
    indexes_to_directions(Indexes,[D1,D2|Directions]).
indexes_to_directions([], _).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: new_direction/4  
%%% INPUT: An index,The list of Indexes
%%%        The list of variables's triples Tertiary
%%% EFFECT: Moving the \theta-constraints on Indexes to the 
%%%        coordinates variables in Tertiary
%%% NOTE: Huge number of NON linear constraints introduced
%%% status: never stable, unless propagation of these
%%%         constraints is possible.
%%% ALE: this is our bottleneck in the constrain framework
%%%      There are other 4 versions (not listed here)
%%%      using more or less refined detail of constraints.
%%%      This is the better version, but it's common
%%%      the great amount of non linear constraints that
%%%      disallow an efficient propagation!
%%% AGO: Domanda da uomo della strada.
%%%      Intuitivamente, perche' usi la base 5 per i pesi? 
%%%      Poiche' gli indici vanno da 0 a 6 mi verrebbe da usare la base 6.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

new_direction([Ax,Ay,Az], [Bx,By,Bz], [Cx,Cy,Cz], I):-
    domain([Rx,Ry,Rz,Tipo,Ax,Ay,Az,Bx,By,Bz,Cx,Cy,Cz],-1,1),
    domain([ANx,ANy,ANz],0,1),
    domain([Nx,Ny,Nz,T1,T2],-2,2),
    domain([I],0,5),
    %% per evitare due vettori consecutivi colineari
    10 * Ax + 5 * Ay + Az #\= 10 *  Bx + 5 * By + Bz,
    %% per evitare vettori a=-b (ridondante, 
     %% perche' c'e' gia' l'alldifferent sulle posizioni)
    10 * Ax + 5 * Ay + Az #\= (-10 *  Bx - 5 * By -Bz),
    %% per evitare vettori con angoli bend <90
    scalar_product([Ax,Ay,Az],[Bx,By,Bz],Ang), Ang #>=0,
    scalar_product([Cx,Cy,Cz],[Bx,By,Bz],Ang1), Ang1 #>=0,
    Ax*Ay*Az #= 0, Ax*Ay + Ax*Az + Ay*Az #\= 0,
    Bx*By*Bz #= 0, Bx*By + Bx*Bz + By*Bz #\= 0,
    Cx*Cy*Cz #= 0, Cx*Cy + Cx*Cz + Cy*Cz #\= 0,
    %%% Relazioni...
    vector_product([Bx,By,Bz],[Ax,Ay,Az],[Nx,Ny,Nz]),
    vector_product([Ax,Ay,Az],[ANx,ANy,ANz],[Rx,Ry,Rz]),
    scalar_product([Rx,Ry,Rz],[Bx,By,Bz],T1),
    scalar_product([ANx,ANy,ANz],[Bx,By,Bz],T2),    
    %%%%%%
    ANx #= 1-Ax*Ax, ANy #= 1-Ay*Ay, ANz #= 1-Az*Az,
    ANx * ANy #=0, ANx * ANz #=0, ANy * ANz #=0, ANx + ANy + ANz #= 1,
    Tipo #= -1*T1*T2,
    type_constraint(I,Tipo,Nx,Ny,Nz,Ax,Ay,Az,Bx,By,Bz,Cx,Cy,Cz).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% PREDICATE: type_constraint
%%%% AGO: predicato da spiegare.

%%%% ALE: this is the key relation between
%%%%      cartesian coordinate system and torsional indexes
%%%% INPUT: Tipo encodes the position of the bend of 90 degrees
%%%%        that influences the type of constraints generated
%%%%  Tipo=0 90 degrees
%%%%  Tipo=1 60 degrees clockwise from the 90 degrees position
%%%%  Tipo=-1 otherwise
%%%% A, B, C are the three bond vectors
%%%% N is the vector = B x A (vect. product)
%%%% OUTPUT: no
%%%% EFFECT: new constraints are added to relate the coordinates of
%%%% four consecutive aa (forming a torsinal angle) to the
%%%% associated torsion index (I)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

type_constraint(I,Tipo,Nx,Ny,Nz,Ax,Ay,Az,Bx,By,Bz,Cx,Cy,Cz) :-
      domain([K1,K2,K3],0,1),
      K1 #= (Tipo+1)/2,  % 
      K2 #= (Tipo+2)/2,  % 
      K3 #= K2-K1,  % 
      (I#=0 #/\ Cx#= -1*Ax+Bx - K3*Bx #/\ 
                 Cy#= -1*Ay+By - K3*By #/\ 
                 Cz#= -1*Az+Bz - K3*Bz     ) #\/
      (I#=1 #/\ (3-K3)*Cx #= (2-K3)*Nx+ (1+K1)*Bx+ (-2+K2)*Ax #/\
                (3-K3)*Cy #= (2-K3)*Ny+ (1+K1)*By+ (-2+K2)*Ay #/\
                (3-K3)*Cz #= (2-K3)*Nz+ (1+K1)*Bz+ (-2+K2)*Az    ) #\/
      (I#=2 #/\ (3-K3)*Cx #= (2-K3)*Nx+ (1-2*K1)*Bx+ (1+K1)*Ax #/\
                (3-K3)*Cy #= (2-K3)*Ny+ (1-2*K1)*By+ (1+K1)*Ay #/\
                (3-K3)*Cz #= (2-K3)*Nz+ (1-2*K1)*Bz+ (1+K1)*Az    ) #\/
      (I#=3 #/\  Cx#=Ax #/\ Cy#=Ay #/\ Cz#=Az)  #\/
      (I#=4 #/\ (-3+5*K2+K1)*Cx #= (2-3*K2-K1)*Nx+ Bx+ (-2+3*K2)*Ax #/\
                (-3+5*K2+K1)*Cy #= (2-3*K2-K1)*Ny+ By+ (-2+3*K2)*Ay #/\
                (-3+5*K2+K1)*Cz #= (2-3*K2-K1)*Nz+ Bz+ (-2+3*K2)*Az    ) #\/
      (I#=5 #/\ (3-1*K2-5*K1)*Cx #= (-2+1*K2+3*K1)*Nx+ (2-1*K2-2*K1)*Bx+ 
                                     (-1+3*K1)*Ax #/\
                (3-1*K2-5*K1)*Cy #= (-2+1*K2+3*K1)*Ny+ (2-1*K2-2*K1)*By+ 
                                     (-1+3*K1)*Ay #/\
                (3-1*K2-5*K1)*Cz #= (-2+1*K2+3*K1)*Nz+ (2-1*K2-2*K1)*Bz+ 
                                     (-1+3*K1)*Az    ).
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: energy_constraints
%%% INPUT: The Primary list of aminoacids,
%%%        The Tertiary list of triples,
%%%        The Indexes list
%%% OUTPUT: Computes in Energy (as a constraint!)
%%%        the value of energy associated to a spatial
%%%        conformation  and the list of potentials
%%% NOTES: Matrix is the NxN matrix containing for 
%%%        each M[i,j] the energy (as constraint) 
%%%        contribution given by the relation 
%%%        between the i-th and j-th aminoacids.
%%%        M is symmetric and thus only the 
%%%        upper triangular part is computed.
%%%        M is represented as M = [ row1, ... , rowN ].
%%%        Potential_List is the list version of M,
%%%        including only non ground elements (to be
%%%        used to predict the not already folded 
%%%        part of protein)
%%% status: to work.
%%% DUBBIO AGO: se facessimo M con assert di fatti m(i,j)
%%%        sarebbe piu' veloce? Da provare.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

energy_constraints(_Primary, Tertiary, _Indexes, 0, _, _) :-
    length(Tertiary, M),
    M =< 6, !.
energy_constraints(Primary, Tertiary, Indexes, Energy, 
    Potential_list,Matrix):- 
    %%% Matrixes creation
    length(Primary,N),
    Matrixlen is N*N,
    length(Matrix,Matrixlen),
    %%%% 
    indexes2segments(Indexes,Segm),

    constrain_matrix(1,Primary, Segm, Tertiary, Matrix, N),
    %%DEBUG: print_matrix(Matrix),
    create_potlist(1,Matrix,N,Potential_list), 

    sum(Potential_list,#=,Energy).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain_matrix
%%% INPUT: ...
%%% I row to analyze
%%% J column to analyze (in the /9 predicate)
%%% A1 A2 the pair of aminoacid contributing for the contact
%%% Segm is the list of ground runs of indexes (to detect areas
%%%      already ground in the matrix)
%%% Tertiary
%%% Matrix
%%% Size of matrix
%%% OUTPUT: only new constraints added (see below for each type)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

constrain_matrix(_,[],_,[],_,_).
constrain_matrix(_,[_],_,[_,_,_],_,_).
constrain_matrix(I, [A1,A2|ListaAmin], Segm, [P1x,P1y,P1z,P2x,P2y,P2z|ListaPos], 
                 Matrix,Size):-
     J is I+1,
     constrain_matrix(I,J,A1,[A2|ListaAmin],Segm,[P1x,P1y,P1z],
                         [P2x,P2y,P2z|ListaPos],Matrix,Size),
     constrain_matrix(J, [A2|ListaAmin], Segm, [P2x,P2y,P2z|ListaPos], 
                         Matrix,Size).

constrain_matrix(_I, _J, _A1,[],_Segm, _, [], _Matrix, _Size).

constrain_matrix(I,J,A1,[A2|ListaAmin],Segm,[P1x,P1y,P1z],
                 [P2x,P2y,P2z|ListaPos],Matrix, Size):-
    J1 is J+1,
    %%% for each pair of aminoacids, apply the constraints on matrix
    constrain_contact(Matrix,Size,I,J,A1,A2,[P1x,P1y,P1z],[P2x,P2y,P2z],Segm),
    constrain_mindist(Matrix,Size,I,J),
    constrain_matrix(I,J1,A1,ListaAmin,Segm,[P1x,P1y,P1z],ListaPos,Matrix,Size).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain_mindist
%%% INPUT: Matrix 
%%%        Size
%%%        I J refers to M[I,J] element (i-th and j-th amino interaction)
%%% NOTE: 3 is the max distance in the sequence that 
%%%       is not generating a contact (excluding i -- i+2)
%%% EFFECT: imposes that in a short range in the sequence (4) 
%%%         the aminoacids have a distance greater than 2 
%%%         (thus they give a null contribution)
%%% status: currently this heuristic is not active
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

constrain_mindist(Matrix,Size,I,J):-
    I+2 < J, J < I+2,!,
    matrixel(Matrix,Size,I,J,0).
constrain_mindist(_Matrix,_Size,_I,_J).
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain_contact
%%% INPUT: Matrix with size NxN
%%%     I J current element M[I,J] referring to the i-th and j-th aminoacids
%%%     A1 A2 amino type ('a','c',...) of i-th and j-th aminoacids
%%%     [X1,Y1,Z1],[X2,Y2,Z2] tertiary positions (constraint variables)
%%%     Segm is the list of ground runs of indexes
%%% EFFECT: 
%%%     if I and J are included in the same ground run,
%%%     then energy is NOT computed and set to be 0 
%%%     (to simplify the search)
%%%     if the aminoacids are in distance less or equal 
%%%     to 2 in the primary sequence, they don't contribute
%%%     in the energy function.
%%%     Otherwise, the constraint binding M[I,J] and 
%%%     the energy contribution is set.
%%%     If the distance (in norm 1) between the two aminoacids 
%%%     is not 2 then the M[I,J]=El is constraint to be 0 else 
%%%     to be the value described in table for the specific 
%%%     type of contact between the two aminoacids.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

constrain_contact(Matrix, N, I, J, _A1, _A2, _, _, Segm):-

      %%% converts I J in indexes
      
      I1 is I-1,
      J1 is J-2,
      included(I1,J1,Segm),!,
      matrixel(Matrix,N,I,J,0).
constrain_contact(Matrix,N,I,J,A1,A2,[X1,Y1,Z1],[X2,Y2,Z2],_Segm):-
 %%% skip contacts for aa with distance of 2 in the sequence
        I + 2 =< J,!,
        matrixel(Matrix,N,I,J,El),
        table(A1, A2, Pot),
        El in {0,Pot},                          
        DX #= abs(X1 - X2),
        DY #= abs(Y1 - Y2),
        DZ #= abs(Z1 - Z2),
        2 #= DX + DY + DZ #<=> El #= Pot.  %%% CONTACT CONTRIBUTION
constrain_contact(Matrix, N, I, J, _A1, _A2, _, _, _):-
      matrixel(Matrix,N,I,J,0).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: create_potlist/4 (aux create_potlist/5)
%%% INPUT: A matrix M, its size, an integer N
%%% OUTPUT: The list PotList of relevant potential
%%% NOTE: It selects from the matrix M, from N to Size,
%%%       only the contribution of non ground
%%%       elements.
%%% status: OK (ALE yes, it should :-) )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

create_potlist(N, Matrix,Size,PotList):-
        N =< Size,!,
        M is N+1,
        create_potlist(N, M, Matrix, Size, List),
        create_potlist(M, Matrix, Size, List1),
        append(List,List1,PotList).
create_potlist(_N, _Matrix, _Size, []).

create_potlist(N, M, Matrix, Size, List):-
        M =< Size,
        matrixel(Matrix,Size,N,M,El),
        ground(El),!,
        %%  El is ground, so is not considered
        M1 is M+1,
        create_potlist(N, M1, Matrix, Size, List).    
create_potlist(N, M, Matrix, Size, [El|List]):-
        M =< Size,!,
        matrixel(Matrix,Size,N,M,El),
        M1 is M+1,
        create_potlist(N, M1, Matrix, Size, List).
create_potlist(_N, _M, _Matrix, _Size, []).
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF CONSTRAINT DEFINITION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% LABELING  PHASE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: solution_search/8 (aux. solutions_search/8)
%%% INPUT: The time limit in seconds (resp. in milliseconds),
%%%        the Primary list of aminoacids, The list of Secondary
%%%        structure Information, The list of Indexes, 
%%%        The flat list Tertiary_flat of coordinates
%%%        the Matrix of contacts, the Frequency list
%%% OUTPUT: The ground values for Tertiary variables!
%%% NOTE: If Indexes are all known, perform an immediate standard labeling.
%%%       Otherwise, look at Indexes, which is a sequence of alternations:
%%%        [(V1),G1,V2,G2,...] of variables and ground values.
%%%          Find the "best" sublist [G1,V2,G2] (or [V1,G1], or [Gn,Vn])
%%%       The best is that having |G1|+|V2|+|G2| - 2 * |V2| bigger
%%%       (i.e. less variables than constant part).
%%%       Perform the labeling on that sublist!!!!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

solution_search(Time,Primary,Secondary,Indexes,Tertiary,Energy,M,Freq):-
     Timemilli is Time * 1000,
     solutions_search(Timemilli,Primary,Secondary,Indexes,Tertiary,Energy,M,Freq).

solutions_search(_Time, _Primary, _Secondary, Indexes, Tertiary, _Energy,_M, _Freq):-
        ground( Indexes ), 
        ground( Tertiary), ! , 
        write('End of solutions search (case I)'),nl.
solutions_search(_Time,  _Primary, _Secondary, Indexes, Tertiary, _Energy,_M, _Freq):-
        ground( Indexes ), ! ,
        write('End of solutions search (case II)'),nl,
%%% NOTA AGO: con 1YPA, 20 secondi qui si pianta in modo anomalo ....
%%% GIRANDO la lista invece funziona. Il tutto ha senso in quanto
%%% se gli indici sono ground e la terziaria no, vuol dire
%%% che i non ground sono all'inizio per via della propagazione
%%% in avanti. Girandola, arriva in fondo. Miracolo.
         reverse(Tertiary,RevTertiary),
         labeling([ff],RevTertiary). 
solutions_search(Time, Primary, Secondary, Indexes,  Tertiary, Energy,Matrix, Freq):-
        length(Indexes,L),
        %%% debugging: 
        write('Solutions search for '), write(L), write(' Indexes:'),nl,
        %%% end debugging.
        Size is L+3,
        create_potlist(1,Matrix,Size,Potential_list),
        choose_labeling(Time, Indexes, Energy, Potential_list, 
                         Tertiary,Primary,Matrix, Freq),
        solutions_search(Time, Primary, Secondary, Indexes, 
                         Tertiary, Energy,Matrix,Freq).
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: choose_labeling/8
%%% INPUT:  The Time in ms, the ... Sublist of the Indexes and the Energy, ...
%%% OUTPUT:  The List of partial potentials, and the associated Energy
%%% NOTE: If the number of vars is little, use built-in labeling.
%%%       Otherwise, use labeling_new (with Heuristics). 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                                              
choose_labeling(_Time,  SubIndexes, Energy, _ , SubTertiary,_SubPrimary,_M, _Freq):-
       %%% (1): built-in
       num_var(SubIndexes, Nvars),
       length(SubIndexes, Sublength),
       %%% Test: (values can be changed)
       Nvars < 4, Sublength < 20,!,
       writetime(video,'Choosen built-in labeling...'), 
       append(SubIndexes,SubTertiary,AA),
       labeling([ff,minimize(Energy)], AA),
       %%% Debugging
       writetime(video,'### built-in labeling end.').
        
choose_labeling( Time,  SubIndexes, Energy, ListaPot, SubTertiary,_SubPrimary, M, Freq):-
        %%% (2): ad hoc (with heuristics)
        write('Chosen ad hoc labeling...'),nl,        
        % assert initial solution and ListaPot
        initial_solution_assert(ListaPot),
        selection_strategy(Selected,Vi,Vt,SubIndexes,SubTertiary,Freq),
%%%%% NOTE: decommenta qui e togli il fail in labeling_new 
%%%%%       per avere versione con minimize
%%%%%%       minimize((
        labeling_new_launch(Time, Selected,Vi,Vt,0, 1, Energy, ListaPot, 
                            M,SubIndexes,SubTertiary,Freq),
%%%%%%     ), Energy ),
        computed_solution_retract.
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% labeling_new_launch/12
%%%%% INPUT: ... tutto ...
%%%%% OUTPUT: ... tutto ...
%%%%% NOTE: time_out/3 is a built-in meta-predicate
%%%%% :- time_out( Goal, Time, L) 
%%%%% 1. returns L = success if a solution for Goal is found in <= Time ms.
%%%%% 2. returns L = time_out if no solution for Goal is found in <= Time ms.
%%%%% 3. fails if in <= Time ms if it is proved that no solutions for Goal exist.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

labeling_new_launch(Time, Selected, Vi,Vt, Done, MOD, 
                    Energy, ListaPot, M,I,Ter,Freq):-
    (time_out(
         labeling_new(Selected, Vi,Vt, Done, MOD, Energy, 
                      ListaPot, M,I,Ter,Freq), Time, _L),
     fail;  %%% FORZATO IL FILE PER PROSEGUIRE SULL'ALBERO !!!!
     best_ind(I)),
     ground(I),!.

labeling_new_launch(_,_,_,_,_,_,_,_,_,_,_,_):-
    nl,write('Error: Not enough time to generate a solution!'),nl,!,fail.

%%% PREDICATE: labeling_new
%%% INPUT: In the call to minimum_domain
%%%      "ff" first-fail, "leftmost" leftmost strategy
%%%  To be better explained
%%% A first-fail or leftmost, self-programmed strategy: 
%%% (2 is the number of variables to be instantiated before
%%%  the next heuristic control)
  
%%% ALE: prima di spiegare tutto aspetterei la  versione nuova (o definitiva...)
%%% mi sa che cambieremo ancora un po qui...

labeling_new(Selected, Vi,Vt, Done, MOD, Energy, ListaPot, M,I,Ter,Freq) :-
       Selected\==done, 
        MOD < 2,
        myindomain(Vi,Vt,Freq),
%%%    write(I),nl,
        Now_Done is Done + 1,
        MOD1 is MOD + 1,
        selection_strategy(Selected1,Vi1,Vt1,I,Ter,Freq),
        labeling_new(Selected1,Vi1,Vt1,Now_Done, MOD1, Energy, ListaPot, M,I,Ter,Freq).

labeling_new(Selected, Vi,Vt, Done, 2, Energy, ListaPot, M,I,Ter,Freq) :-
      Selected\==done, 
      myindomain(Vi,Vt,Freq),
      best_listapot(BestPL), 
      term_variables(ListaPot,LpV),
      length(LpV,LVarPot),
      length(ListaPot,LTotPot),
      sumgrounds( ListaPot, BestPL, Valint, ValintBest),
      Now_Done is Done + 1,
      CoefS1 is (0.5 * (LVarPot) + 1.1 * (LTotPot-LVarPot)) / (LTotPot),
      S1  is integer( CoefS1 * ValintBest)+1,


      Valint  < S1,

%      write(Valint), write(' '), write(S1), write(' '), 
%      write(ValintBest), write(' '),write( LVarPot),
%        writetime(' '),
%        write(I),nl,

      selection_strategy(Selected1,Vi1,Vt1,I,Ter,Freq),
      labeling_new(Selected1,Vi1,Vt1,Now_Done, 1, Energy, ListaPot, M,I,Ter,Freq).

labeling_new(done, _,_, _,  _, Energy, ListaPot, _M, I,Ter,_Freq) :-
        best_solution( Val ),
        %%%write(Ter),nl,
        %%%write(I),nl,
        Energy < Val, 
  %%% Debugging
        writetime(video,'### ad-hoc labeling end in time: '),nl,
        %%%print_results('pippo',SubPrimary,SubTertiary),
        %%%write(ListaPot),nl,
        %%%write(I),nl,
        %%%write(Energy),nl,
        computed_solution_retract,
        update_solution(Energy,ListaPot,Ter,I),!,
%%%%%%%%%%%%%% togli il fail e decommenta la minimize per avere la versione vecchia
          fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% indomain redefined %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %%%%%%%
%%% PREDICATE: myindomain/3   (aux. mydomaini)
%%% INPUT: Two variables, one of them instantiated to "no", 
%%%        and the Frequency list for the variables.
%%% OUTPUT: A possible value, computed by 
%%% standard clpfd domain, or
%%% mydomaini, that uses frequency order.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

myindomain(A,B,_Freq):-
   A==no,!,
   indomain(B).   
myindomain(A,B,Freq):-
   B==no,!,
   myindomaini(A,Freq).   

myindomaini(V,Freq):-
    assoc_lookup(Freq,V,Order),
    member(V,Order).

%%% AGO: bello sarebbe non avere accesso sequenziale a
%%% Freq qui. Provare a pensarci....
%%% NOTA: Se mettiamo anche le probabilita' degli 
%%% angoli torsionali, bisogna cambiare qui.

assoc_lookup([[Var,Order]|_],Var1, Order):-
    Var == Var1.
assoc_lookup([[Var,_]|R],Var1,Order):-
   Var \== Var1,
   assoc_lookup(R,Var1,Order).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% selection_strategy(Status,SelectedIndex,SelectedTertiary,Indexes,Tertiary,_Freq)
%%% Status: done/notdone
%%% SelectedIndex,SelectedTertiary: unified with no/variable selected
%%%  selection_strategy(notdone,Vindex,Vtertiary,_Indexes,Tertiary,_Freq):-

%%% if some tertiary not labelled (due to failed propagation)
%%% -> applied on left propag. of indexes on tertiary

%%% NOTE: Fattorizzato reverse e findVG per non ripetere
%%% l'operazione.

selection_strategy(notdone,Sel1,Sel2,Indexes,Tertiary,_Freq):-
    reverse(Tertiary,Ter_rev),
    findVG(Ter_rev,P),
    ( Sel1 = no,
      P3 is (P-1)//3,
      nth(P3,Indexes,El),
      ground(El),
      nth(P,Tertiary,Sel2),!; %%% 1
      Sel2 = no,
      P3 is (P-1)//3+1,
      split(Indexes,P3,[Sel1|_R]),!;%%% 2
      Sel1 = no,
      term_variables_bag(Ter_rev,List),
      List\==[],!,
      List=[Sel2|_R] ). %%% 3

selection_strategy(done,no,no,_Indexes,_Tertiary,_Freq).

%%% P is the position of the rightmost Var el with a Ground el at the right
%
% _,1,_,2,_,3,4,5,_
% 1 2 3 4 5 6 7 8 9 
%
% P=5
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
findVG(List,P):-
     vars(List,Nv,L2),
     grounds(L2,Ng,_),
     length(List,L),
     P is L-Nv-Ng.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

split(List,Pos,Nlist):-
    length(Temp,Pos),
    prefix(Temp,List),
    term_variables(Temp,TempV),
    length(List,L),
    P1 is L-Pos,
    length(Temp1,P1),
    suffix(Temp1,List),
    term_variables(Temp1,Temp1V),
    reverse(TempV,Temp2V),
    merge(Temp2V,Temp1V,Nlist).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% SUBLIST EXTRACTION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% PREDICATE: choose_subindexes (aux. admissible_sub, best_sub)
%%% INPUT: The list of Indexes,                                        
%%% OUTPUT: Two indexes Start and Length (starting point and
%%% length of the sum of variables and non variables (3 alternations)
%%% at the beginning

choose_subindexes(Indexes, Start, Length, Secondary):-
        admissible_sub(Indexes, 1, ListaPoss),
        %%%write('Possibilities: '),write(ListaPoss),nl,
        best_sub(ListaPoss, Start, Length, Secondary,Indexes).        

%%% PREDICATE: admissible_sub
%%% INPUT: The list of Indexes, a number I where
%%%        the analysis of Indexes starts
%%% OUTPUT: A list of 4-tuples
%%%      [ K,
%%%        [L1,L2,L3,L4,L5,L6], 
%%%      { where K is an entry in Indexes, 
%%%       L1 is the length of non variables starting from K
%%%       L2 is the length of the successive variables,    
%%%       L3 is the length of successive non variables,
%%%       ...
%%%       L6 is the  length of the successive variables
%%%       }
%%%       L = L1 + ... + L6,
%%%       #vars in the sublist Indexes(K,K + L -1)
%%%      ]
%%% Example: :- admissible_sub(
%%%     [1,2,A,B,C,3,4,D,E,F,5,6,G,H,2,3,4,5,6,7,D,F],1,L).
%%% returns:
%%% L = [[1,[2,3,2,3,2,2],14,8],
%%%      [6,[2,3,2,2,6,2],15,5],
%%%      [11,[2,2,6,2,0,0],12,4]] 
%%% Observe that K starts a non-ground sequence, if possible.
%%% status: critical point.
%%% NOTE:further cases can be introduced to optimize!!!!

admissible_sub(Indexes, 1, [[1, [0,Nvars1,L2,Nvars2,L3,Nvars3], Length, Nvars]|ListaPoss]):-
       vars(Indexes, Nvars1, R1 ),
        Nvars1 > 0, !,
       grounds(R1, L2, R2),
       vars(R2, Nvars2, R3 ),
       grounds(R3, L3, R4 ),
       vars(R4, Nvars3, _R5 ),
         (%%Case 1: [ Vars ]
          R1 = [], !,      
          ListaPoss = [],
          Length = Nvars1,
          Nvars = Nvars1,
          L2 = 0;

          %%Case 2: [ Vars, Grounds | ...]
          R2=[],!,
          Nvars=Nvars1, 
          Length is L2 + Nvars1,
          NextStart is Nvars1 + 1,
          ListaPoss = [];

          %%Case 2/bis: [ Vars, Grounds, Vars| ...]
          R3=[],!,
          Nvars is Nvars1+Nvars2,
          Length is L2 + Nvars1 + Nvars2,
          NextStart is Nvars1 + 1,
          admissible_sub(R1, NextStart, ListaPoss);

          %%Case 2/ter: [ Vars, Grounds, Vars, Grounds| ...]
          R4=[],!,         
          Nvars is Nvars1+Nvars2,
          Length is L2 + Nvars1 + Nvars2+L3,
          NextStart is Nvars1 + 1,
          admissible_sub(R1, NextStart, ListaPoss);

          %%Case 2/q: [ Vars, Grounds, Vars, Grounds| ...]
          
          Nvars is Nvars1+Nvars2+Nvars3,
          Length is L2 + Nvars1 + Nvars2+L3+Nvars3,
          NextStart is Nvars1 + 1,
          admissible_sub(R1, NextStart, ListaPoss)  
          ).

admissible_sub(Indexes, Start, [[Start,[L1,Nvars1,L2,Nvars2,L3,Nvars3],Length, Nvars]| ListaPoss]):-
        grounds(Indexes, L1, R1),
         L1 > 0, !,
         vars(R1,Nvars1,R2),
         grounds(R2,L2,R3),
         vars(R3,Nvars2,R4),
         grounds(R4,L3,R5),
         vars(R5,Nvars3,R6),
         (%%% Case 3: [ Grounds ]
          R1 = [], !, 
          ListaPoss = [],
          Nvars = 0,
          Length is Nvars + L1;
          
          %%% Case 4: [ Grounds , Vars ]
          R2 = [], !, 
          ListaPoss = [],
          Nvars = Nvars1,
          Length is Nvars + L1;
          
          %%% Case 5: [ Ground , Vars , Ground]
          R3 = [], !, 
          ListaPoss = [],
          Nvars = Nvars1,
          Length is  Nvars1 + L1 + L2;
          %%% Case6: [ Grounds , Vars , Grounds , Vars ]         
             R4 = [], !,  %
         ListaPoss = [],
          Nvars is Nvars1 + Nvars2,
          Length is L1 + Nvars1 + L2 + Nvars2;

          %%% Case7:  [ Grounds , Vars , Ground , Vars , Ground ]        
          R5 = [], !, 
         ListaPoss = [],
         Length is L1 + Nvars1 + L2 + Nvars2 + L3,
          Nvars is Nvars1 + Nvars2;

          %%%% Case8:  [ Grounds , Vars , Ground , Vars , Ground | ...]
          R6 = [], !, 
         Length is L1 + Nvars1 + L2 + Nvars2 + L3,
          Nvars is Nvars1 + Nvars2,
          NewStart is Start + L1 + Nvars1,
          admissible_sub(R2,NewStart,ListaPoss);

          %%%% Case9:  [ Grounds , Vars , Ground , Vars , Ground, V | ...]
         Length is L1 + Nvars1 + L2 + Nvars2 + L3 + Nvars3,
          Nvars is Nvars1 + Nvars2 + Nvars3,
          NewStart is Start + L1 + Nvars1,
          admissible_sub(R2,NewStart,ListaPoss)).
%%% Case 3: Indexes = [ Ground ]
admissible_sub(_, _, []).

        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: best_sub/3 (aux. best_subaux/3)
%%% INPUT: A list of triples: [Starting index, L , Nvar ]
%%%       identifying sublists [G1,V1,G2,V2,G3,V3] where
%%%       L = |G1|+|V1|+|G2|+ |V2| + |G3| + |V3| , 
%%%       Nvar = |V1| + |V2| + |V3| 
%%%       (cf previous admissible_sub predicate's description).
%%% OUTPUT: The best tuple (the last value is changed,
%%%      for optimization)
%%% NOTE: We maximize L - 2 * Nvar
%%%       We use best_sub/3 for optimizing
%%%       If more than 10 variables are in the selected sublist,
%%%       we cut it, using possible_cut
%%% NOTE: The check includes the presence of ssbonds, that
%%%       increase the value of a sublist.
%%% status: critical point.

best_sub(List, Start, Length, Secondary ,I) :-
       best_subaux(List, [_,_,_,-1000,_],Tuple, Secondary,I),
       %%%%% write(Tuple),nl,
       possible_cut(Tuple, Start, Length, Secondary).

%%%if included a ssbond the score is increased by 100 
best_subaux([ [Start,[C1,V1,C2,V2,C3,V3],L,Nvar] | R], [_,_,_,Pbest,_],Triple, Secondary,I) :-
       check_include_ssbond(Start,L,SS, Secondary,I,A,B),
%%% if less than 10 variables, set NV=10
       NV #=10 #<=> Nvar #=< 10,
       NV #=Nvar #<=> Nvar#> 10,
       P1 is L - 2*NV+100*SS,
       %%%%%% write(Start),write(' '),write(L),write(' '),
         %%%%%% write(SS),write(' '),write(P1),write(' '),write(Pbest),nl,
       P1 > Pbest,!,
       best_subaux( R , [Start, [C1,V1,C2,V2,C3,V3], L, P1,[A,B]] ,Triple, Secondary,I).
best_subaux([ [_,_,_,_] | R],[StartBest, Qbest,Lbest,NvarBest,Bond],Triple, Secondary,I) :-
        best_subaux( R , [StartBest, Qbest,Lbest,NvarBest,Bond] ,Triple, Secondary,I).
best_subaux([] , BestTriple,BestTriple, _Secondary,_).

possible_cut([Start,[C1,V1,C2,V2,C3,V3],Len,_,[A,B]], S, L, _Secondary) :-
      V1+V2+V3 =<12,
       A>0,
       B>0,!,
      Ai is A-1,
      Bi is B-2,

      %begin
      (
        Ai<Start+C1,!,
        S is Start;

    Ai<Start+C1+V1,!,
        S is Ai;

        Ai<Start+C1+V1+C2,!,
        S is Start+C1+V1;
        
        Ai<Start+C1+V1+C2+V2,!,
        S is Ai;

        S is Start+C1+V1+C2+V2

      ),

      %end
      (
    Bi<Start+C1+V1,!,
        End is Bi;

        Bi<Start+C1+V1+C2,!,
        End is Start+C1+V1+C2-1;
        
    Bi<Start+C1+V1+C2+V2,!,
        End is Bi;

    Bi<Start+C1+V1+C2+V2+C3,!,
        End is Start+Len-1;

        End is Bi

      ),

      L is End-S+1.

possible_cut([Start,[C1,V1,C2,V2,C3,V3],_L,_,_], Start, L1, _Secondary) :-
%%%write(V1),write(' '),
%%%write(V2),write(' '),
%%%write(V3),write(' '),
%%% v3 excluded!
        V1 + V2 + V3 #=< 10,
         !,
        L1 is C1+V1+C2+V2+C3.
possible_cut([Start,[C1,V1,C2,V2,C3,_V3],_,_,_], Start, L, _Secondary) :-
         V1 +V2 #=< 13,!,
         L is C1 + V1 + C2+V2+C3.
possible_cut([Start,[C1,V1,C2,_V2,_C3,_V3],_,_,_], Start, L, _Secondary) :-
       V1 #=< 13,!,
         L is C1 + V1 + C2,
         write('### weak sublist cut'),nl.
possible_cut([Start,[C1,V1,C2,V2,C3,_V3],_,_,_], Start1, L, _Secondary) :-
       V2 #=< 13,!,
         Start1 is Start+C1+V1,
         L is C2 + V2 + C3,
         write('### weak sublist cut'),nl.
possible_cut([Start,[C1,_V1,_C2,_V2,_C3],_,_,_], Start, L, _Secondary) :-
         L is C1 + 13,
         write('### hard sublist cut'),nl.

%%% model
%%% 1 2 3 4 5 6 7  triples 3d
%%%    1 2 3 4 5   indexes
%%% SS = 1 if at least one var included in the tuple by the ssbond(A,B)

check_include_ssbond(Start,L,SS, [ssbond(A,B)|_R],Ind,A,B):-
%% Ai and Bi are the corresponding on indexes
   Ai is A-1,
   Bi is B-2,
   Len is Bi-Ai+1,

   Start =< Ai,
   Bi =< Start+L-1, %%if not included cut
%% ssbond included in tuple
%% at least one var included?
   sublist(Ai,Len,Ind,Subind),
   num_var(Subind,Nvar),
   Nvar >= 1,!,
   SS is 1.
check_include_ssbond(Start,L,SS, [ssbond(_A,_B)|R],Ind,A,B):-
   check_include_ssbond(Start,L,SS, R,Ind,A,B).
check_include_ssbond(_Start,_L,0, _,_,-1,-1).

%%%%%%%%%% END MAIN CODE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% AUXILIARY (TECHNICAL AND EASY) PREDICATES %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: checkamino/1
%%% INPUT: A list
%%% OUTPUT: True iff the list contains aminoacids
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

checkamino([T|R]):-
    ground(T),
    aminoacid(T),!,
    checkamino(R).
checkamino([_T|_R]):-
    write('Bad aminoacid sequence!.'),nl,
    fail.
checkamino([]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PREDICATE: terelem/3
%% INPUT: the number A of one aminoacid and the
%%        Tertiary structure.
%% OUTPUT: The coordinates [X,Y,Z] corresponding to amino A 
%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

terelem(Tertiary,A ,[X,Y,Z]):-
    Ind11a is (A-1)*3+1, 
    Ind12a is Ind11a + 1, 
    Ind13a is Ind12a + 1,
    nth(Ind11a,Tertiary,X), 
    nth(Ind12a,Tertiary,Y), 
    nth(Ind13a,Tertiary,Z).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: matrixel
%%% INPUT: A Matrix, its size N, numbers I and J.
%%% OUTPUT: M[I,J] if there is such a value. A new unbounded
%%%         variable otherwise.
%%% status: Capire perche' unbounded value e non 0.
%%%         ALE:: 0 va bene, gli accessi non sono mai 
%%%               nella zona esterna. 
%%%               (il predicato e' solo per completezza e magari 
%%%               si puo' proprio togliere per controllare errori...)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

matrixel(Matrix,N,I,J,El):-
   0 < I, I =< N,
   0 < J, J =< N,
   I < J,!,
   Ind is J+(I-1)*N,
   nth(Ind, Matrix, El).
%matrixel(_Matrix,_N,_I,_J,0).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: indexes2segments/2 (aux. indexes2segments/3)
%%% INPUT: a list of ground and nonground indexes, e.g.
%%%       indexes2segments([X,Y,0,0,0,A,B,C,1,1,1],L).            
%%% OUTPUT: The intervals where finding ground sublists, e.g.
%%%       L = [[3,5],[9,11]]. 
%%% NOTE: Cambiato nella terza clausola. Spero di non aver fatto danni.
%%% status: OK
%%%         controllato: ho sistemato i "danni" in energy_constraint
%%%         ora e' usato correttamente la'.
%%%         Anche avoidsymmetries va bene!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

indexes2segments(Indexes,R):-
      indexes2segments(Indexes,R,1).

indexes2segments(Indexes,R,P):-
       vars(Indexes, Nv, I1),
       Nv>0,!,
       P1 is P + Nv,
       indexes2segments(I1,R,P1).
indexes2segments(Indexes,[[A,B]|R1],P):-
       grounds(Indexes, Ng, I1),
       Ng >0,!,
       A is P, %%AGO: cambiato%A is P+1,
       B is P + Ng - 1, %%%B is P+Ng+1,
       P1 is P+Ng,
       indexes2segments(I1,R1,P1).
indexes2segments([],[],_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PREDICATE: getmaxrun (aux getmax)
%% INPUT: A (sub)list of Indexes 
%% OUTPUT: A pointer S to the beginning of the longest
%%         ground sequence of the protein.
%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

getmaxrun(Indexes, S):-
    indexes2segments(Indexes, Runs),
    getmax(Runs,0,0,S).

getmax([],S,_,S).
getmax([[S,E]|R],S1,E1,Q):-
     E-S > E1-S1,!,
     getmax(R,S,E,Q).
getmax([_|R],Me,Ms,Q):-
     getmax(R,Me,Ms,Q).   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATES: vector_sum, vector_minus, 
%%%             vector_product, scalar_product
%%% INPUT: V1, V2 (ecp. vector_minus)
%%% OUTPUT: Z = V1 vector_operation V2.
%%% NOTE1: vector_product and scalar_product introduce
%%%       NON linear constraints 
%%% NOTE2: internal (clpfd:scalar_product/4) cannot be used
%%%        with the first argument non instantiated.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

vector_minus([X,Y,Z],[X1,Y1,Z1]):-
        X1 #= 0 - X,
        Y1 #= 0 - Y,
        Z1 #= 0 - Z.
vector_sum([A,B,C],[D,E,F],[G,H,I]):-
        G #= A + D,
        H #= B + E,
        I #= C + F.
vector_product([Ax,Ay,Az],[Bx,By,Bz],[Cx,Cy,Cz]):-
        Cx #= Ay*Bz - Az*By,
        Cy #= Az*Bx - Ax*Bz,
        Cz #= Ax*By - Ay*Bx.
scalar_product([Ax,Ay,Az],[Bx,By,Bz],S):-
        Px #= Ax*Bx,
          Py #= Ay*By,
          Pz #= Az*Bz,
        sum([Px,Py,Pz],'#=',S).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: merge
%%% merges two lists.
%%% status: stable, of course
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge([A|R],[B|R1],[A,B|R2]):-
     merge(R,R1,R2).
merge([],R1,R1).
merge(R1,[],R1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: included
%%% states if the interval [A,B] is included in one of the
%%% intervals [Ai,Bi] in the list [[A1,B1],[A2,B2], ..., [An, Bn]]
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

included(A,B,[[A1,B1]|_R]):-
     A1 =< A,
     B =< B1,!.
included(A,B,[ _ |R]):-
     included(A,B,R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: num_var.
%%% INPUT: a list L
%%% OUTPUT: The number of non ground variables in the list L
%%% NOTE: It uses the built-in term_variables 
%%%       (returns the list of variables)
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

num_var(L,N):-
   term_variables(L,Varlist),
   length(Varlist,N).

%%% vars([A,B,C,a,b], X , Y )
%%% returns X = 3 (number of vars at the beginning) and
%%%         Y = [a,b] (the list but the first initial vars)

vars([],0,[]).
vars([X|Y], 0, [X|Y]):-
        ground(X).
vars([X|Y], L, R):-
        var(X),
        vars(Y, M, R),
        L is M+1.  

%%% grounds([a,b,c,A,B,C], X , Y )
%%% returns X = 3 (number of non vars at the beginning) and
%%%         Y = [A,B,C] (the list bust the first initial non vars)

grounds([],0,[]).
grounds([X|Y], 0, [X|Y]):-
        var(X).
grounds([X|Y], L, R):-
        ground(X),
        grounds(Y, M, R),
        L is M+1.  

%%% sumgrounds([2,X,3,C,5,1],C )
%%% returns C = 11 (the sum of the non ground values)

sumgrounds( [ A | R ] , [_B|Best], S, S1 ) :-
         var(A), !,
         sumgrounds( R , Best, S, S1 ). 
sumgrounds( [ A | R ] , [B|Best], S, S1 ) :-
         sumgrounds( R , Best, Si, Si1 ),
         S is A + Si,
         S1 is B + Si1.
sumgrounds( [], [], 0, 0).


sumgrounds( [ A | R ] , C ) :-
         var(A), !,
         sumgrounds( R , C ). 
sumgrounds( [ A | R ] , C  ) :-
         sumgrounds( R , C1 ),
         C is A + C1.
sumgrounds( [], 0).


sumgroundster( [ A,B,C | R ] , [Rx,Ry,Rz], T ) :-
   sumgroundster( [ A,B,C | R ] , [Rx,Ry,Rz], T1 ,N ),
   T is T1 /N.
sumgroundster( [ A,B,C | R ] , [Rx,Ry,Rz], T , N) :-
         ground(A), 
         ground(B), 
         ground(C), !,
         sumgroundster(R,[Rx,Ry,Rz], T1 ,N1),
         N is N1+1,
         T is T1 + abs(Rx-A)+abs(Ry-B)+abs(Rz-C).
sumgroundster( [ _ , _, _ | R ] , Ref, T ,N ) :-
         sumgroundster( R , Ref, T ,N).
sumgroundster( [], _ , 0,0).



%%% sublist(2,3,[a,b,c,d,e,f],X) 
%%% returns X = [b,c,d] (the list of 3 elements starting from 2)

sublist(_Inizio, 0,[],[]).
sublist(1, N, List, Sublist):-
        length( Sublist, N),
        prefix( Sublist, List).
sublist(Inizio, N, [_Testa|Coda], Sublist):-
        Inizio > 1,
        NuovoInizio is Inizio-1,
        sublist( NuovoInizio, N, Coda, Sublist).

%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE even.
%%% status: stable, of course!
%%%%%%%%%%%%%%%%%%%%%%%%%%%

even(X) :- 
    0 #= X mod 2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The list of the 20 aminoacids  %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

aminoacid(a). aminoacid(c). aminoacid(d).
aminoacid(e). aminoacid(f). aminoacid(g).
aminoacid(h). aminoacid(i). aminoacid(k).
aminoacid(l). aminoacid(m). aminoacid(n).
aminoacid(p). aminoacid(q). aminoacid(r). 
aminoacid(s). aminoacid(t). aminoacid(v).
aminoacid(w). aminoacid(y). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Conversion short names/full names %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
sn_fn(a,'ALA'). sn_fn(c,'CYS'). sn_fn(d,'ASP').
sn_fn(e,'GLU'). sn_fn(f,'PHE'). sn_fn(g,'GLY').
sn_fn(h,'HIS'). sn_fn(i,'ILE'). sn_fn(l,'LEU').
sn_fn(m,'MET'). sn_fn(n,'ASN'). sn_fn(p,'PRO').
sn_fn(q,'GLN'). sn_fn(r,'ARG'). sn_fn(s,'SER').
sn_fn(t,'THR'). sn_fn(v,'VAL'). sn_fn(w,'TRP').
sn_fn(y,'TYR'). sn_fn(k,'LYS').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% PRINT PREDICATES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% According to PDB Format Description Version 2.2 
%%% http://www.rcsb.org/pdb/docs/format/pdbguide2.2/guide2.2_frame.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_results(ID, Time, Primary, Secondary, Tertiary,Compact) :-
    system:working_directory(DIR,DIR),
    name(DIR,ListDir),
    append(ListDir,[47,111,117,116,112,117,116,45],Temp1),

    name(ID,ProtL),
    append(Temp1,ProtL,Temp2),
    append(Temp2,[46,112,100,98],Fulllist),


    name(FilePath,Fulllist),
    open(FilePath,write,File),
    print_results(File, ID, Time, Primary, Secondary, Tertiary,Compact),
    close(File),
    print_results(video,ID, Time, Primary, Secondary, Tertiary,Compact).

print_results(MEDIA,ID, Time, Primary, Secondary, Tertiary,Compact) :-
     %%% first line
     stampa(MEDIA,'HEADER      '),
     stampa(MEDIA,ID), 
     stampa(MEDIA,'\n'),
     %%% initial lines
     stampa(MEDIA,'REMARK      '),
     stampa(MEDIA,'Input search time limit: '),
     stampa(MEDIA,Time),
     stampa(MEDIA,' s.   '),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'REMARK      '),
     stampa(MEDIA,'Compact factor: '),
     stampa(MEDIA,Compact),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'REMARK      '),
     writetime(MEDIA,'Global time: '),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'REMARK      '),
     write_global_energy(MEDIA,Primary,Tertiary),
     %%% secondary information
     print_secondary(MEDIA,Primary,Secondary),
     %%% output model
     stampa(MEDIA,'MODEL       1'),
     stampa(MEDIA,'\n'),
     %%% Calpha coordinates
     print_results_rec(MEDIA, 1 , Primary, Tertiary ),
     %%% ending lines
     stampa(MEDIA,'ENDMDL'),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'END'),
     stampa(MEDIA,'\n').

print_results_rec(MEDIA, I , [Amino|R], [X,Y,Z|N]) :-
    stampa(MEDIA,'ATOM    '), %%% 1--6
    (I>99;I>9,stampa(MEDIA,' ');I>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,I), %% 7-8 (atom number)
    stampa(MEDIA,'  CA  '), %%% 9-17
    sn_fn(Amino,AminoName), %% 18-20
    stampa(MEDIA,AminoName), 
    stampa(MEDIA,' A '), %% 21-23
    (I>99;I>9,stampa(MEDIA,' ');I>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,I), %% 24-25  (residue number)
    stampa(MEDIA,'     '), %% 26-30
    Xr is integer(X*3.8/sqrt(2)*1000),
    Yr is integer(Y*3.8/sqrt(2)*1000),
    Zr is integer(Z*3.8/sqrt(2)*1000),
    Xrd is Xr/1000,
    (Xrd>99.999;Xrd>9,stampa(MEDIA,' ');Xr>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,Xrd),
    (Xrr is Xr mod 10, Xrr=0, stampa(MEDIA,'0');
     Xrr1 is Xr mod 100, Xrr1=0, stampa(MEDIA,'00');
     Xrr2 is Xr mod 1000, Xrr2=0, stampa(MEDIA,'000');
      true), 
    stampa(MEDIA,' '), %%31-38
    Yrd is Yr/1000,
    (Yrd>99.999;Yrd>9,stampa(MEDIA,' ');Yr>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,Yrd),
    (Yrr is Yr mod 10, Yrr=0, stampa(MEDIA,'0');
     Yrr1 is Yr mod 100, Yrr1=0, stampa(MEDIA,'00');
     Yrr2 is Yr mod 1000, Yrr2=0, stampa(MEDIA,'000');
     true), 
    stampa(MEDIA,' '), %%39-46
    Zrd is Zr/1000,
    (Zrd>99.999;Zrd>9,stampa(MEDIA,' ');Zr>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,Zrd),
    (Zrr is Zr mod 10, Zrr=0, stampa(MEDIA,'0');
     Zrr1 is Zr mod 100, Zrr1=0, stampa(MEDIA,'00');
     Zrr2 is Zr mod 1000, Zrr2=0, stampa(MEDIA,'000');
     true), 
    stampa(MEDIA,'  '), %%47-54
    stampa(MEDIA,'1.00  1.00'),      
    stampa(MEDIA,'\n'),
    I1 is I + 1,
    print_results_rec(MEDIA, I1, R, N).
print_results_rec(_, _ , [], []).

print_secondary(MEDIA,Primary,[ssbond(_,_)|R]):-
    !,%%% Per ora niente per gli ssbond
    print_secondary(MEDIA,Primary,R).
print_secondary(MEDIA,Primary,R):-
    print_helix(MEDIA,Primary,1,R).
print_helix(MEDIA,Primary,N,[ helix(A,B)|R]) :-
    !,
    stampa(MEDIA,'HELIX  '),%1-7
    (N>99;N>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,N), %8-11
    (N>99;N>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,'H'),stampa(MEDIA,N),stampa(MEDIA,' '), %12-15
    nth(A,Primary,Amino),
    sn_fn(Amino,Aname),
    stampa(MEDIA,Aname), stampa(MEDIA,' '), %16-19
    stampa(MEDIA,N),stampa(MEDIA,'  '), %20-21
    (A>99;A>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,A),
    stampa(MEDIA,'  '), %22-27
    nth(B,Primary,AminoB),
    sn_fn(AminoB,Bname),
    stampa(MEDIA,Bname), stampa(MEDIA,' '), %28-31
    stampa(MEDIA,N),stampa(MEDIA,'  '), %32-33
    (B>99;B>9,stampa(MEDIA,' ');B>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,B), 
    stampa(MEDIA,'  1'),
    stampa(MEDIA,'\n'), %34-
    M is N + 1,
    print_helix(MEDIA,Primary,M,R).
print_helix(MEDIA,Primary,_N,R) :-
    print_strand(MEDIA,Primary,1,R).
print_strand(MEDIA,Primary,N, [ strand(A,B)|R]) :-
    !,
    stampa(MEDIA,'SHEET  '),%1-7
    (N>99;N>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,N), stampa(MEDIA,'   A'), %8-11
    stampa(MEDIA,' 1 '), %12-14
    nth(A,Primary,Amino),
    sn_fn(Amino,Aname),
    stampa(MEDIA,Aname), stampa(MEDIA,' A'), %18-21
    stampa(MEDIA,' '), %22
    (A>99;A>9,stampa(MEDIA,' ');A>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,A),
    stampa(MEDIA,'  '), %22-27
    nth(B,Primary,AminoB),
    sn_fn(AminoB,Bname),
    stampa(MEDIA,Bname), stampa(MEDIA,' A'), %28-31
    stampa(MEDIA,' '), %32-33
    (B>99;B>9,stampa(MEDIA,' ');B>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,B), 
    stampa(MEDIA,'  0                                        '),
    stampa(MEDIA,'\n'), %34-
    M is N + 1,
    print_strand(MEDIA,Primary,M,R).
print_strand(_,_,_,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% stampa/2 
%%% INPUT: a Stream or the string 'video' and a String
%%% EFFECT: It prints the String in the stream (a file)
%%%         or in the video
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stampa(video,String) :-
   write(String).
stampa(Stream,String) :-
   write(Stream,String).
   
%%%%%%%%%%%%%%%%

print_matrix(M,I):-
    length(M,LM),
    N is integer(sqrt(LM)),
    print_matrix(M,N,N,[*,*,*|I]).

print_matrixt(M,I,T):-
    length(M,LM),
    N is integer(sqrt(LM)),
    print_matrix(M,N,N,[*,*,*|I],T).

print_matrix(M):-
    length(M,LM),
    N is integer(sqrt(LM)),
    print_matrix(M,N,N).

print_matrix(_Matrix,0,_N).
print_matrix(Matrix,M,N):-
   A is (N-M)*N+1,
   Chi is (N-M)+1,
   sublist(A,N,Matrix,List),
   write(Chi),write(' '),
   print_line_matrix(List),nl,
   M1 is M-1,
   print_matrix(Matrix,M1,N).

print_matrix(_Matrix,0,_N,_).
print_matrix(Matrix,M,N,[Ii|I]):-
   A is (N-M)*N+1,
   Chi is (N-M)+1,
   sublist(A,N,Matrix,List),
   write(Chi),write(' '),
   print_line_matrix(List),
   write('|'),write(Ii),nl,
   M1 is M-1,
   print_matrix(Matrix,M1,N,I).

print_matrix(_Matrix,0,_N,_,_).
print_matrix(Matrix,M,N,[Ii|I],[T1,T2,T3|Tr]):-
   A is (N-M)*N+1,
   Chi is (N-M)+1,
   sublist(A,N,Matrix,List),
   write(Chi),write(' '),
   print_line_matrix(List),
   write('|'),write(Ii),
   write(' |'),write(T1),
   write(' '),write(T2),
   write(' '),write(T3),nl,
   M1 is M-1,
   print_matrix(Matrix,M1,N,I,Tr).

print_line_matrix([]).
print_line_matrix([A|R]):-
   var(A),!,
   write('.'),write(','),
   print_line_matrix(R).
print_line_matrix([A|R]):-
   A=0,!,
   write(A),write(','),
   print_line_matrix(R).
print_line_matrix([_|R]):-
   write('*,'),
   print_line_matrix(R).

print_matrixd(M):-
    length(M,LM),
    N is integer(sqrt(LM)),
    print_matrixd(M,N,N).
print_matrixd(_Matrix,0,_N).
print_matrixd(Matrix,M,N):-
    A is (N-M)*N+1,
    Chi is (N-M)+1,
    sublist(A,N,Matrix,List),
    write(Chi),write(' '),
    print_line_matrixd(List),nl,
    M1 is M-1,
    print_matrixd(Matrix,M1,N).
   
print_line_matrixd([]).
print_line_matrixd([A|R]):-
    fd_max(A,AA),
    fd_min(A,AA1),
    fd_size(A,AA2),
    write(AA),write('-'),
    write(AA1),write('!'),
    write(AA2),write(','),
    print_line_matrixd(R).

printtable:-
    table(A,B,C),
    write('table('),write(A),write(' ,'),write(B),write(', '),write(C),write(').'),nl,
    fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: write_global_energy/3
%%% EFFECT: It writes the free energy of the folding computed 
%%%         according Fgolari et al. model. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_global_energy(MEDIA, Primary,Tertiary) :-
    potential_list(Primary,Tertiary,List),
    sum(List,#=,Energy),
    stampa(MEDIA,'Energy: '),
    stampa(MEDIA, Energy),
    stampa(MEDIA,'\n').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: potential_list/3 (aux. lista_pot_amino)
%%% This predicate considers all contacts for Tertiary.
%%% It allows to compute the final Energy according to
%%% Fogolari's papers. 
%%% Currently it is used for the final output only,
%%% to compare the results with those of WFLP02 and ERCIM03.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

potential_list([],[],[]).
potential_list([_],[_,_,_],[]).
potential_list( [A1,A2|ListaAmin], [P1x,P1y,P1z,P2x,P2y,P2z|ListaPos], Lnew):-
        potential_list([A2|ListaAmin], [P2x,P2y,P2z|ListaPos], L),
        list_pot_amino(A1,[P1x,P1y,P1z],ListaAmin,ListaPos,Lnew,L).

list_pot_amino(Amino,[X1,Y1,Z1],[Amino2|ListaAmin],[X2,Y2,Z2|ListaPos],[C|L],L1):-
        table(Amino, Amino2, Pot),
        C in {0,Pot},
        DX #= abs(X1 - X2),
        DY #= abs(Y1 - Y2),
        DZ #= abs(Z1 - Z2),
        2 #= DX + DY + DZ  #<=> C #= Pot,  %%% CONTACT CONTRIBUTION
        list_pot_amino(Amino,[X1,Y1,Z1],ListaAmin,ListaPos,L,L1).
list_pot_amino(_Amino,_Posizione,[],[],L,L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% CLEANING PREDICATES
%%% For removing side effects
%%% caused by execution abort
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clean :-  
   retract(start(_)),
   clean.
clean :-
   retract(best_solution(_)),
   clean.
clean :-
   retract(best_ter(_)),
   clean.
clean :-
   retract(best_ind(_)),
   clean.
clean :-
   retract(best_listapot(_)),
   clean.
clean.

%%%% AGO: Perche' questo e' separato dagli altri?
%%%% Non si puo' mettere tutto in clean?
%%%% E' anche chiamato dalla procedura sotto a cui pero'
%%%% visto il nome, farei fare solo l'assert...

clean_bestlistapot :-
   retract(best_listapot(_)),
   clean_bestlistapot.
clean_bestlistapot.

initial_solution_assert(ListaPot) :-
      length(ListaPot,P),
      length(M0,P),
      domain(M0,0,0),
      assert( best_solution(10000) ),
      assert( best_ind(_) ),
      assert( best_ter(_) ),
      clean_bestlistapot,
      assert( best_listapot(M0) ).
      
computed_solution_retract :-
      retract(best_solution(_)),
      retract(best_ter(_)),
      retract(best_ind(_)),
      retract(best_listapot(_)).

update_solution(Energy,ListaPot,Ter,I) :-
      assert( best_solution(Energy)),
      assert( best_listapot(ListaPot)),
      assert( best_ter(Ter)),
      assert( best_ind(I)).
      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% Timing Predicates: %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

initialization :-
    clean,      %%% remove garbage if any
    statistics(runtime,[T|_]), 
    assert(start(T)).

writetime(MEDIA,Str) :-
    start(InitialTime),
    statistics(runtime,[CurrentTime|_]),
    Delta is CurrentTime - InitialTime,
    stampa(MEDIA,Str),
    print_time(MEDIA,Delta),
    stampa(MEDIA,'\t').

print_time(MEDIA, K) :-
    K < 60000,!,
    S is K//1000,
    M is K mod 1000,
    stampa(MEDIA,S),
    stampa(MEDIA,','),
    stampa(MEDIA,M),
    stampa(MEDIA,'s').
print_time(MEDIA, K ) :-
  K < 3600000,!,
  M is K //60000,
  S is K mod 60000,
  stampa(MEDIA,M),
  stampa(MEDIA,'min.'),
  print_time(MEDIA,S).
print_time(MEDIA, K ) :-
  H is K //3600000,
  M is K mod 3600000,
  stampa(MEDIA,H),
  stampa(MEDIA,'hours.'),
  print_time(MEDIA,M).
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%END%%
